import re
import sys
import pytest

from unittest.mock import patch
from typing import Dict
from pathlib import Path
from bs4 import BeautifulSoup
from sphinx.cmd.build import main as sphinx_main

from ._check import check_link_targets

# Insert test package root to path for all tests
sys.path.insert(0, str(Path(__file__).parent / "src"))

default_conf = """
extensions = [
    "sphinx.ext.autodoc",
    "sphinx_codeautolink",
]

autodoc_default_options = {
    "members": True,
    "undoc-members": True,
}
"""

any_whitespace = re.compile(r'\s*')
ref_tests = list(Path(__file__).with_name('ref').glob('*.txt'))
ref_xfails = {
    'ref_fluent_attrs.txt': sys.version_info < (3, 8),
    'ref_fluent_call.txt': sys.version_info < (3, 8),
    'ref_import_from_complex.txt': sys.version_info < (3, 8),
}


def assert_links(file: Path, links: list):
    text = file.read_text('utf-8')
    soup = BeautifulSoup(text, 'html.parser')
    blocks = list(soup.find_all('a', attrs={'class': 'sphinx-codeautolink-a'}))
    strings = [any_whitespace.sub('', ''.join(b.strings)) for b in blocks]

    assert len(strings) == len(links)
    for s, link in zip(strings, links):
        assert s == link


@pytest.mark.parametrize('file', ref_tests)
def test_references(file: Path, tmp_path: Path):
    """
    Basic extension tests for reference building.

    The tests are structured as .txt files, parsed and executed here.
    The structure of the file is::

        expected
        autolink
        link.targets
        # split
        lines to add to the default conf.py
        # split
        index.html content
    """
    if ref_xfails.get(file.name, False):
        pytest.xfail('Expected to fail.')

    links, conf, index = file.read_text('utf-8').split('# split')
    links = links.strip().split('\n')
    if len(links) == 1 and not links[0]:
        links = []

    files = {'conf.py': default_conf + conf, 'index.rst': index}
    result_dir = _sphinx_build(tmp_path, 'html', files)

    assert_links(result_dir / 'index.html', links)
    assert check_link_targets(result_dir) == len(links)


table_tests = list(Path(__file__).with_name('table').glob('*.txt'))


@pytest.mark.parametrize('file', table_tests)
def test_tables(file: Path, tmp_path: Path):
    """
    Tests for backreference tables.

    The tests are structured as .txt files, parsed and executed here.
    The structure of the file is::

        expected
        table
        link.targets
        # split
        lines to add to the default conf.py
        # split
        index.html content

    Note that the header of the table is also considered a link target.
    However, if the table is collapsible, the header is not a part of
    the table, so it should be omitted from the expected links.
    The processing also removes any whitespace, which should be taken into account.
    """
    links, conf, index = file.read_text('utf-8').split('# split')
    links = links.strip().split('\n')
    if len(links) == 1 and not links[0]:
        links = []

    files = {'conf.py': default_conf + conf, 'index.rst': index}
    result_dir = _sphinx_build(tmp_path, 'html', files)

    index_html = result_dir / 'index.html'
    text = index_html.read_text('utf-8')
    soup = BeautifulSoup(text, 'html.parser')
    blocks = list(soup.select('table a'))
    strings = [any_whitespace.sub('', ''.join(b.strings)) for b in blocks]

    assert len(strings) == len(links)
    for s, link in zip(strings, links):
        assert s == link


fail_tests = list(Path(__file__).with_name('fail').glob('*.txt'))


@pytest.mark.parametrize('file', fail_tests)
def test_fails(file: Path, tmp_path: Path):
    """
    Tests for failing builds.

    The tests are structured as .txt files, parsed and executed here.
    The structure of the file is::

        lines to add to the default conf.py
        # split
        index.html content
    """
    conf, index = file.read_text('utf-8').split('# split')
    files = {'conf.py': default_conf + conf, 'index.rst': index}
    with pytest.raises(RuntimeError):
        _sphinx_build(tmp_path, 'html', files)


def test_non_html_build(tmp_path: Path):
    index = """
Test project
------------

.. code:: python

   import test_project
   test_project.bar()

.. automodule:: test_project
.. autolink-examples:: test_project.bar
"""
    files = {'conf.py': default_conf, 'index.rst': index}
    _sphinx_build(tmp_path, 'man', files)


def test_build_twice_and_modify_one_file(tmp_path: Path):
    index = """
Test project
------------

.. code:: python

   import test_project
   test_project.bar()

.. automodule:: test_project

.. toctree::

   another
"""
    another = """
Another
-------
.. autolink-examples:: test_package.bar
"""
    another2 = """
Another
-------
But edited.

.. autolink-examples:: test_package.bar
"""
    files = {'conf.py': default_conf, 'index.rst': index, 'another.rst': another}
    _sphinx_build(tmp_path, 'html', files)
    _sphinx_build(tmp_path, 'html', {'another.rst': another2})


def test_build_twice_and_delete_one_file(tmp_path: Path):
    index = """
Test project
------------

.. code:: python

   import test_project
   test_project.bar()

.. automodule:: test_project

.. toctree::

   another
"""
    another = """
Another
-------
.. autolink-examples:: test_project.bar
"""

    files = {'conf.py': default_conf, 'index.rst': index, 'another.rst': another}
    _sphinx_build(tmp_path, 'html', files)
    (tmp_path / 'src' / 'another.rst').unlink()
    _sphinx_build(tmp_path, 'html', {})


def test_raise_unexpected(tmp_path: Path):
    index = """
Test project
------------

.. code:: python

   import test_project
   test_project.bar()

.. automodule:: test_project
"""
    files = {'conf.py': default_conf, 'index.rst': index}

    def raise_msg(*args, **kwargs):
        raise ValueError('ValueError')

    def raise_nomsg(*args, **kwargs):
        raise ValueError()

    target = 'sphinx_codeautolink.extension.CodeBlockAnalyser'
    with pytest.raises(RuntimeError), patch(target, raise_msg):
        _sphinx_build(tmp_path, 'html', files)

    with pytest.raises(RuntimeError), patch(target, raise_nomsg):
        _sphinx_build(tmp_path, 'html', files)


def test_parallel_build(tmp_path: Path):
    index = """
Test project
------------
.. automodule:: test_project
.. toctree::
"""
    template = """
{header}
---
.. code:: python

   import test_project
   test_project.bar()
"""
    links = ['test_project', 'test_project.bar']

    n_subfiles = 20
    subfiles = {
        name: template.format(header=name)
        for name in map(lambda x: f'F{x}', range(n_subfiles))
    }
    index = index + '\n   '.join([''] + list(subfiles))
    files = {'conf.py': default_conf, 'index.rst': index}
    files.update({k + '.rst': v for k, v in subfiles.items()})
    result_dir = _sphinx_build(tmp_path, 'html', files, n_processes=4)

    for file in subfiles:
        assert_links(result_dir / (file + '.html'), links)
    assert check_link_targets(result_dir) == n_subfiles * len(links)


def _sphinx_build(
    folder: Path, builder: str, files: Dict[str, str], n_processes: int = None
) -> Path:
    """Build Sphinx documentation and return result folder."""
    src_dir = folder / 'src'
    src_dir.mkdir(exist_ok=True)
    for name, content in files.items():
        (src_dir / name).write_text(content, 'utf-8')

    build_dir = folder / 'build'
    args = ['-M', builder, str(src_dir), str(build_dir), '-W']
    if n_processes:
        args.extend(['-j', str(n_processes)])
    ret_val = sphinx_main(args)
    if ret_val:
        raise RuntimeError('Sphinx build failed!')
    return build_dir / builder



import sphinx_codeautolink


class TestPackage:
    def test_version(self):
        assert sphinx_codeautolink.__version__

    def test_clean_pycon_public(self):
        assert sphinx_codeautolink.clean_pycon

    def test_clean_ipython_public(self):
        assert sphinx_codeautolink.clean_ipython



import pytest

from ._util import refs_equal, skip_walrus


class TestAssign:
    @refs_equal
    def test_assign_before_import(self):
        s = 'a = 1\nimport a\na'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_assign_after_import(self):
        s = 'import a\na = 1\na'
        refs = [('a', 'a')]
        return s, refs

    @refs_equal
    def test_assign_to_other_name_linked(self):
        s = 'import a\nb = a'
        refs = [('a', 'a'), ('a', 'a'), ('a', 'b')]
        return s, refs

    @refs_equal
    def test_assign_uses_and_assigns_imported(self):
        s = 'import a\na = a\na'
        refs = [('a', 'a'), ('a', 'a'), ('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_unpack_assign_starred(self):
        s = '*a, b = c'
        refs = []
        return s, refs

    @refs_equal
    def test_unpack_assign_uses_and_overwrites(self):
        s = 'import a\na, b = a\na'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_multilevel_unpack_assign_uses_and_overwrites(self):
        s = 'import a\n(a, b), c = a\na'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_multitarget_assign_to_different_names(self):
        s = 'import a\nc = b = a'
        refs = [('a', 'a'), ('a', 'a'), ('a', 'b'), ('a', 'c')]
        return s, refs

    @refs_equal
    def test_multitarget_assign_uses_and_overwrites(self):
        s = 'import a\na = b = a\na, b'
        refs = [('a', 'a'), ('a', 'a'), ('a', 'b'), ('a', 'a'), ('a', 'a'), ('a', 'b')]
        return s, refs

    @refs_equal
    def test_multitarget_assign_overwrites_twice(self):
        s = 'import a\na = a = a'
        refs = [('a', 'a'), ('a', 'a'), ('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_multitarget_assign_overwrites_twice_with_attribute(self):
        s = 'import a\na = a = a.b'
        refs = [('a', 'a'), ('a.b', 'a.b'), ('a.b', 'a'), ('a.b', 'a')]
        return s, refs

    @refs_equal
    def test_multitarget_assign_with_unpacking_first_assigns_to_other(self):
        s = 'import a\n(c, d) = b = a'
        refs = [('a', 'a'), ('a', 'a'), ('a', 'b')]
        return s, refs

    @refs_equal
    def test_multitarget_assign_with_unpacking_last_assigns_to_other(self):
        s = 'import a\nd = (b, c) = a'
        refs = [('a', 'a'), ('a', 'a'), ('a', 'd')]
        return s, refs

    @refs_equal
    def test_assign_uses_and_assigns_modified_imported(self):
        s = 'import a\na = a + 1\na'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_assign_subscript_uses(self):
        s = 'import a\na.b[0] = 1'
        refs = [('a', 'a'), ('a.b', 'a.b')]
        return s, refs

    @refs_equal
    def test_assign_call_subscript_uses(self):
        s = 'import a\na.b()[0] = 1'
        refs = [('a', 'a'), ('a.b', 'a.b')]
        return s, refs

    @refs_equal
    def test_augassign_uses_imported(self):
        s = 'import a\na += 1\na'
        refs = [('a', 'a'), ('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_augassign_uses_and_assigns_imported(self):
        s = 'import a\na += a\na'
        refs = [('a', 'a'), ('a', 'a'), ('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_annassign_overwrites_imported(self):
        s = 'import a\na: b = 1\na'
        refs = [('a', 'a')]
        return s, refs

    @refs_equal
    def test_annassign_uses_and_assigns_imported(self):
        s = 'import a\nb: 1 = a\nb.c'
        refs = [('a', 'a'), ('a', 'a'), ('a', 'b'), ('a.c', 'b.c')]
        return s, refs

    @refs_equal
    def test_annassign_uses_and_annotates_imported(self):
        s = 'import a\nb: a = 1\nb.c'
        refs = [('a', 'a'), ('a', 'a'), ('a.()', 'b'), ('a.().c', 'b.c')]
        return s, refs

    @refs_equal
    def test_annassign_prioritises_annotation(self):
        s = 'import a, b\nc: a = b\nc.d'
        # note that AnnAssign is executed from value -> annot -> target
        refs = [
            ('a', 'a'),
            ('b', 'b'),
            ('b', 'b'),
            ('a', 'a'),
            ('a.()', 'c'),
            ('a.().d', 'c.d'),
        ]
        return s, refs

    @refs_equal
    def test_annassign_why_would_anyone_do_this(self):
        s = 'import a\na: a = a\na.b'
        refs = [('a', 'a'), ('a', 'a'), ('a', 'a'), ('a.()', 'a'), ('a.().b', 'a.b')]
        return s, refs

    @refs_equal
    def test_annassign_without_value_overrides_annotation_but_not_linked(self):
        # note that this is different from runtime behavior
        # which does not overwrite the variable value
        s = 'import a\na: b\na'
        refs = [('a', 'a')]
        return s, refs

    @skip_walrus
    @refs_equal
    def test_walrus_uses_imported(self):
        s = 'import a\n(a := 1)\na'
        refs = [('a', 'a')]
        return s, refs

    @skip_walrus
    @refs_equal
    def test_walrus_uses_and_assigns_imported(self):
        s = 'import a\n(a := a)\na'
        refs = [('a', 'a'), ('a', 'a'), ('a', 'a'), ('a', 'a')]
        return s, refs

    @skip_walrus
    @refs_equal
    def test_walrus_uses_and_assigns_modified_imported(self):
        s = 'import a\n(a := a + 1)\na'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @skip_walrus
    @refs_equal
    def test_nested_walrus_statements(self):
        s = 'import a\n(c := (b := a))'
        refs = [('a', 'a'), ('a', 'a'), ('a', 'b'), ('a', 'c')]
        return s, refs

    @skip_walrus
    @refs_equal
    def test_walrus_result_assigned(self):
        s = 'import a\nc = (b := a)'
        refs = [('a', 'a'), ('a', 'a'), ('a', 'b'), ('a', 'c')]
        return s, refs

    @refs_equal
    def test_dotted_import_overwrites_all_components(self):
        s = 'class a:\n  b = 1\nimport a.b\na.b'
        refs = [('a.b', 'a.b'), ('a.b', 'a.b')]
        return s, refs

    @pytest.mark.xfail(reason='Following assigns into imports would be a pain.')
    @refs_equal
    def test_partially_overwrite_dotted_import(self):
        s = 'import a.b.c\na.b = 1\na\na.b\na.b.c'
        refs = [('a.b.c', 'a.b.c'), ('a', 'a')]
        return s, refs


class TestFollowAssignment:
    @refs_equal
    def test_follow_simple_assign(self):
        s = 'import a\nb = a\nb'
        refs = [('a', 'a'), ('a', 'a'), ('a', 'b'), ('a', 'b')]
        return s, refs

    @refs_equal
    def test_follow_simple_assign_attr(self):
        s = 'import a\nb = a\nb.c'
        refs = [('a', 'a'), ('a', 'a'), ('a', 'b'), ('a.c', 'b.c')]
        return s, refs

    @refs_equal
    def test_follow_attr_assign(self):
        s = 'import a\nc = a.b\nc'
        refs = [('a', 'a'), ('a.b', 'a.b'), ('a.b', 'c'), ('a.b', 'c')]
        return s, refs

    @refs_equal
    def test_follow_attr_assign_attr(self):
        s = 'import a\nc = a.b\nc.d'
        refs = [('a', 'a'), ('a.b', 'a.b'), ('a.b', 'c'), ('a.b.d', 'c.d')]
        return s, refs

    @refs_equal
    def test_follow_attr_call_assign_attr(self):
        s = 'import a\nc = a.b()\nc.d'
        refs = [('a', 'a'), ('a.b', 'a.b'), ('a.b.()', 'c'), ('a.b.().d', 'c.d')]
        return s, refs

    @refs_equal
    def test_follow_attr_call_assign_attr_call(self):
        s = 'import a\nc = a.b()\nc.d()'
        refs = [('a', 'a'), ('a.b', 'a.b'), ('a.b.()', 'c'), ('a.b.().d', 'c.d')]
        return s, refs

    @refs_equal
    def test_follow_through_two_complex_assignments(self):
        s = 'import a\nd = a.b().c\nf = d().e'
        refs = [
            ('a', 'a'),
            ('a.b', 'a.b'),
            ('a.b.().c', 'c'),
            ('a.b.().c', 'd'),
            ('a.b.().c', 'd'),
            ('a.b.().c.().e', 'e'),
            ('a.b.().c.().e', 'f'),
        ]
        return s, refs


class TestAssignLike:
    @refs_equal
    def test_with_uses_imported(self):
        s = 'import a\nwith a as b:\n  a'
        refs = [('a', 'a'), ('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_with_overwrites_imported(self):
        s = 'import a\nwith 1 as a:\n  a'
        refs = [('a', 'a')]
        return s, refs

    @refs_equal
    def test_with_uses_and_overwrites_imported(self):
        s = 'import a\nwith a as a:\n  a'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_for_uses_imported(self):
        s = 'import a\nfor b in a:\n  a'
        refs = [('a', 'a'), ('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_for_overwrites_imported(self):
        s = 'import a\nfor a in b:\n  a'
        refs = [('a', 'a')]
        return s, refs

    @refs_equal
    def test_for_uses_and_overwrites_imported(self):
        s = 'import a\nfor a in a:\n  a'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_async_for_uses(self):
        s = 'import a\nasync def f():\n  async for b in a:\n    a'
        refs = [('a', 'a'), ('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_for_else_uses(self):
        s = 'import a\nfor b in c:\n  pass\nelse:\n  a'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs



"""Directive implementations."""
from docutils import nodes
from docutils.parsers.rst import Directive, directives
from sphinx import addnodes


class DeferredExamples(nodes.Element):
    """Deferred node for substitution later when references are known."""

    def __init__(self, ref: str, collapse: bool):
        super().__init__()
        self.ref = ref
        self.collapse = collapse

    def copy(self):
        """Copy element."""
        return self.__class__(self.ref, self.collapse)


class Examples(Directive):
    """Gather and display references in code examples."""

    has_content = False
    required_arguments = 1
    optional_arguments = 0
    option_spec = {
        'collapse': directives.flag,
        'type': directives.unchanged,
    }

    def run(self):
        """Run directive to insert a :class:`DeferredExamples`."""
        name = self.arguments[0]
        collapse = self.options.get('collapse', False) is None
        par = nodes.paragraph()
        deferred = DeferredExamples(name, collapse)
        par += deferred
        ref = addnodes.pending_xref(
            refdomain='py',
            refexplicit=False,
            refwarn=False,
            reftype=self.options.get('type', 'class'),
            reftarget=name,
        )
        ref += nodes.literal(classes=['xref', 'py', 'py-class'], text=name)
        deferred += ref
        return [par]


class ConcatMarker(nodes.Element):
    """Marker for :class:`Concat`."""

    def __init__(self, mode: str = None):
        super().__init__()
        self.mode = mode

    def copy(self):
        """Copy element."""
        return self.__class__(self.mode)


class Concat(Directive):
    """Toggle and cut literal block concatenation in a document."""

    has_content = False
    required_arguments = 0
    optional_arguments = 1

    def run(self):
        """Insert :class:`ConcatMarker`."""
        arg = self.arguments[0] if self.arguments else 'on'
        return [ConcatMarker(arg)]


class PrefaceMarker(nodes.Element):
    """Marker for :class:`Preface`."""

    def __init__(self, content: str):
        super().__init__()
        self.content = content

    def copy(self):
        """Copy element."""
        return self.__class__(self.content)


class Preface(Directive):
    """Include a preface in the next code block."""

    has_content = True
    required_arguments = 0
    optional_arguments = 1
    final_argument_whitespace = True

    def run(self):
        """Insert :class:`PrefaceMarker`."""
        lines = list(self.arguments) + list(self.content)
        return [PrefaceMarker('\n'.join(lines))]


class SkipMarker(nodes.Element):
    """Marker for :class:`Skip`."""

    def __init__(self, level: str):
        super().__init__()
        self.level = level

    def copy(self):
        """Copy element."""
        return self.__class__(self.level)


class Skip(Directive):
    """Skip auto-linking next code block."""

    has_content = False
    required_arguments = 0
    optional_arguments = 1

    def run(self):
        """Insert :class:`SkipMarker`."""
        arg = self.arguments[0] if self.arguments else 'next'
        return [SkipMarker(arg)]


class RemoveExtensionVisitor(nodes.SparseNodeVisitor):
    """Silently remove all codeautolink directives."""

    def unknown_departure(self, node):
        """Ignore unknown nodes."""

    def unknown_visit(self, node):
        """Remove nodes."""
        if isinstance(node, DeferredExamples):
            # Remove surrounding paragraph too
            node.parent.parent.remove(node.parent)
            return
        elif isinstance(node, (ConcatMarker, PrefaceMarker, SkipMarker)):
            node.parent.remove(node)



"""Resolve import locations and type hints."""
from dataclasses import dataclass
from functools import lru_cache
from importlib import import_module
from inspect import isclass, isroutine
from typing import Optional, Tuple, Any, Union, Callable, List
from ..parse import Name, NameBreak


def resolve_location(chain: Name, inventory) -> str:
    """Find the final type that a name refers to."""
    comps = []
    cursor = None
    for comp in chain.import_components:
        if comp != NameBreak.call:
            comps.append(comp)
            continue

        if cursor is None:
            comps, cursor = make_cursor(comps)

        cursor = locate_type(cursor, tuple(comps), inventory)
        if cursor is None:
            raise CouldNotResolve()

        call_value(cursor)
        comps = []

    if cursor is None:
        try:
            comps, cursor = make_cursor(comps)
        except CouldNotResolve:
            # Last ditch effort to locate based on the string only
            return '.'.join(comps)

    cursor = locate_type(cursor, tuple(comps), inventory)
    return cursor.location if cursor is not None else None


class CouldNotResolve(Exception):
    """Could not resolve type to inventory."""


@dataclass
class Cursor:
    """Cursor to follow imports, attributes and calls to the final type."""

    location: str
    value: Any
    instance: bool


def make_cursor(components: List[str]) -> Tuple[List[str], Cursor]:
    """Divide components into module and rest, create cursor for following the rest."""
    value, index = closest_module(tuple(components))
    if value is None or index is None:
        raise CouldNotResolve()
    location = '.'.join(components[:index])
    return components[index:], Cursor(location, value, False)


def locate_type(
    cursor: Cursor, components: Tuple[str, ...], inventory
) -> Cursor:
    """Find type hint and resolve to new location."""
    previous = cursor
    for i, component in enumerate(components):
        cursor = Cursor(
            cursor.location + '.' + component,
            getattr(cursor.value, component, None),
            cursor.instance,
        )

        if cursor.value is None:
            raise CouldNotResolve()

        if isclass(cursor.value):
            cursor.instance = False

        if (
            isclass(cursor.value)
            or (isroutine(cursor.value) and cursor.location not in inventory)
        ):
            # Normalise location of type or imported function
            try:
                cursor.location = fully_qualified_name(cursor.value)
            except (AttributeError, TypeError):
                # Odd construct encountered: don't try to be clever but continue
                pass

        if isclass(previous.value) and cursor.location not in inventory:
            for val in previous.value.mro():
                name = fully_qualified_name(val)
                if name + '.' + component in inventory:
                    previous.location = name
                    return locate_type(previous, components[i:], inventory)

        previous = cursor

    return cursor


def call_value(cursor: Cursor):
    """Call class, instance or function."""
    if isclass(cursor.value) and not cursor.instance:
        # class definition: "instantiate" class
        cursor.instance = True
        return

    if callable(cursor.value) and not isroutine(cursor.value):
        # callable class instance
        cursor.value = cursor.value.__call__
    elif not isroutine(cursor.value):
        raise CouldNotResolve()  # not a function either

    cursor.value = get_return_annotation(cursor.value)
    cursor.location = fully_qualified_name(cursor.value)
    cursor.instance = True


def get_return_annotation(func: Callable) -> Optional[type]:
    """Determine the target of a function return type hint."""
    annotations = getattr(func, '__annotations__', {})
    ret_annotation = annotations.get('return', None)

    # Inner type from typing.Optional or Union[None, T]
    origin = getattr(ret_annotation, '__origin__', None)
    args = getattr(ret_annotation, '__args__', None)
    if origin is Union and len(args) == 2:
        nonetype = type(None)
        if args[0] is nonetype:
            ret_annotation = args[1]
        elif args[1] is nonetype:
            ret_annotation = args[0]

    # Try to resolve a string annotation in the module scope
    if isinstance(ret_annotation, str):
        location = fully_qualified_name(func)
        mod, _ = closest_module(tuple(location.split('.')))
        ret_annotation = getattr(mod, ret_annotation, ret_annotation)

    if (
        not ret_annotation
        or not isinstance(ret_annotation, type)
        or hasattr(ret_annotation, '__origin__')
    ):
        raise CouldNotResolve()

    return ret_annotation


def fully_qualified_name(thing: Union[type, Callable]) -> str:
    """Construct the fully qualified name of a type."""
    return thing.__module__ + '.' + thing.__qualname__


@lru_cache(maxsize=None)
def closest_module(components: Tuple[str, ...]) -> Tuple[Any, Optional[int]]:
    """Find closest importable module."""
    try:
        mod = import_module(components[0])
    except ImportError:
        return None, None

    for i in range(1, len(components)):
        try:
            mod = import_module('.'.join(components[:i + 1]))
        except ImportError:
            # import failed, exclude previously added item
            return mod, i
    # imports succeeded, include all items
    return mod, len(components)



import pytest
from sphinx_codeautolink.parse import Component
from ._util import refs_equal


class TestUnit:
    def test_component_from_unrecognised_ast(self):
        with pytest.raises(ValueError):
            Component.from_ast('not ast')


class TestSimple:
    @refs_equal
    def test_empty_source(self):
        return '', []

    @refs_equal
    def test_no_imports(self):
        return '1\na = 2\nb()', []

    @refs_equal
    def test_builtins(self):
        s = 'print()'
        refs = [('print', 'print')]
        return s, refs

    @pytest.mark.xfail(reason='Magics are currently not tracked.')
    @refs_equal
    def test_magics(self):
        s = '__file__'
        refs = [('__file__', '__file__')]
        return s, refs

    @refs_equal
    def test_import_from(self):
        s = 'from lib import a'
        refs = [('lib', 'lib'), ('lib.a', 'a')]
        return s, refs

    @refs_equal
    def test_import_from_as(self):
        s = 'from lib import a as b'
        refs = [('lib', 'lib'), ('lib.a', 'a')]
        return s, refs

    @refs_equal
    def test_import_from_multiline(self):
        s = 'from lib import (\n  a,\n  b,\n)'
        refs = [('lib', 'lib'), ('lib.a', 'a'), ('lib.b', 'b')]
        return s, refs

    @refs_equal
    def test_import_from_as_multiline(self):
        s = 'from lib import (\n  a as b,\n  c as d,\n)'
        refs = [('lib', 'lib'), ('lib.a', 'a'), ('lib.c', 'c')]
        return s, refs

    @refs_equal
    def test_simple_import_then_access(self):
        s = 'import lib\nlib'
        refs = [('lib', 'lib'), ('lib', 'lib')]
        return s, refs

    @refs_equal
    def test_inside_list_literal(self):
        s = 'import lib\n[lib]'
        refs = [('lib', 'lib'), ('lib', 'lib')]
        return s, refs

    @refs_equal
    def test_inside_subscript(self):
        s = 'import lib\n0[lib]'
        refs = [('lib', 'lib'), ('lib', 'lib')]
        return s, refs

    @refs_equal
    def test_outside_subscript(self):
        s = 'import lib\nlib[0]'
        refs = [('lib', 'lib'), ('lib', 'lib')]
        return s, refs

    @refs_equal
    def test_simple_import_then_attrib(self):
        s = 'import lib\nlib.attr'
        refs = [('lib', 'lib'), ('lib.attr', 'lib.attr')]
        return s, refs

    @refs_equal
    def test_import_as_then_attrib(self):
        s = 'import lib as b\nb.attr'
        refs = [('lib', 'lib'), ('lib.attr', 'b.attr')]
        return s, refs

    @refs_equal
    def test_import_from_then_attrib(self):
        s = 'from lib import a\na.attr'
        refs = [('lib', 'lib'), ('lib.a', 'a'), ('lib.a.attr', 'a.attr')]
        return s, refs

    @refs_equal
    def test_import_from_as_then_attrib(self):
        s = 'from lib import a as b\nb.attr'
        refs = [('lib', 'lib'), ('lib.a', 'a'), ('lib.a.attr', 'b.attr')]
        return s, refs

    @refs_equal
    def test_dotted_import(self):
        s = 'import a.b\na.b'
        refs = [('a.b', 'a.b'), ('a.b', 'a.b')]
        return s, refs

    @refs_equal
    def test_dotted_import_then_only_part(self):
        s = 'import a.b\na'
        refs = [('a.b', 'a.b'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_dotted_import_then_attrib(self):
        s = 'import a.b\na.b.c'
        refs = [('a.b', 'a.b'), ('a.b.c', 'a.b.c')]
        return s, refs

    @refs_equal
    def test_relative_import_is_noop(self):
        s = 'from .a import b\nb'
        refs = []
        return s, refs

    @refs_equal
    def test_del_removes_import(self):
        s = 'import a\ndel a\na'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_del_dotted_removes_only_part(self):
        s = 'import a.b\ndel a.b\na'
        refs = [('a.b', 'a.b'), ('a.b', 'a.b'), ('a', 'a')]
        return s, refs

    @pytest.mark.xfail(reason='Assignments to imports are not tracked.')
    @refs_equal
    def test_overwrite_dotted_not_tracked(self):
        s = 'import a.b\na.b = 1\na.b.c'
        refs = [('a.b', 'a.b')]
        return s, refs

    @refs_equal
    def test_import_star(self):
        s = 'from sphinx_codeautolink import *\nsetup'
        refs = [
            ('sphinx_codeautolink', 'sphinx_codeautolink'),
            ('sphinx_codeautolink.setup', 'setup')
        ]
        return s, refs



import pytest
from ._util import refs_equal, skip_match


@skip_match
class TestMatch:
    @refs_equal
    def test_match_link_nothing(self):
        s = 'match a:\n  case b(c, d=e):\n    pass'
        refs = []
        return s, refs

    @refs_equal
    def test_match_target_linked(self):
        s = 'import a\nmatch a:\n  case _:\n    pass'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_case_class_linked(self):
        s = 'import a\nmatch _:\n  case a.C():\n    pass'
        refs = [('a', 'a'), ('a.C', 'a.C')]
        return s, refs

    @refs_equal
    def test_case_inner_attribute_linked(self):
        s = 'import a\nmatch _:\n  case a.C(attr=0):\n    pass'
        refs = [('a', 'a'), ('a.C', 'a.C'), ('a.C.().attr', 'attr')]
        return s, refs

    @refs_equal
    def test_case_inner_pattern_linked(self):
        s = 'import a\nmatch _:\n  case C(a.b):\n    pass'
        refs = [('a', 'a'), ('a.b', 'a.b')]
        return s, refs

    @refs_equal
    def test_case_inner_kw_pattern_linked(self):
        s = 'import a\nmatch _:\n  case C(attr=a.b):\n    pass'
        refs = [('a', 'a'), ('a.b', 'a.b')]
        return s, refs

    @refs_equal
    def test_case_inner_kw_pattern_class_linked(self):
        s = 'import a\nmatch _:\n  case C(attr=a.C()):\n    pass'
        refs = [('a', 'a'), ('a.C', 'a.C')]
        return s, refs

    @refs_equal
    def test_case_inner_pattern_with_single_name_overrides(self):
        s = 'import a\nmatch _:\n  case C(a):\n    a'
        refs = [('a', 'a')]
        return s, refs

    @refs_equal
    def test_case_inner_pattern_overrides_but_able_to_use_simultaneously(self):
        s = 'import a\nmatch _:\n  case C(a, a.b):\n    a'
        refs = [('a', 'a'), ('a.b', 'a.b')]
        return s, refs

    @refs_equal
    def test_case_nested_class_linked(self):
        s = 'import a\nmatch _:\n  case C(attr=[a.D()]):\n    pass'
        refs = [('a', 'a'), ('a.D', 'a.D')]
        return s, refs

    @refs_equal
    def test_case_class_attr_target_linked(self):
        s = 'import a\nmatch _:\n  case a.C(attr=x):\n    pass'
        refs = [
            ('a', 'a'), ('a.C', 'a.C'), ('a.C.().attr', 'attr'), ('a.C.().attr', 'x')
        ]
        return s, refs

    @refs_equal
    def test_case_kw_pattern_overrides(self):
        s = 'import a\nmatch _:\n  case a.C(attr=a):\n    a'
        refs = [
            ('a', 'a'),
            ('a.C', 'a.C'),
            ('a.C.().attr', 'attr'),
            ('a.C.().attr', 'a'),
            ('a.C.().attr', 'a'),
        ]
        return s, refs

    @pytest.mark.xfail(reason='Match overriding not implemented.')
    @refs_equal
    def test_case_pattern_overrides_but_able_to_use_simultaneously(self):
        s = 'import a\nmatch _:\n  case a.C(a, attr=a.b):\n    a'
        refs = [('a', 'a'), ('a.C', 'a.C'), ('a.C.().attr', 'attr'), ('a.b', 'a.b')]
        return s, refs

    @pytest.mark.xfail(reason='Match overriding not implemented.')
    @refs_equal
    def test_case_kw_pattern_overrides_but_able_to_use_simultaneously(self):
        s = 'import a\nmatch _:\n  case a.C(attr=a, bttr=a.b):\n    pass'
        refs = [
            ('a', 'a'),
            ('a.C', 'a.C'),
            ('a.C.().attr', 'attr'),
            ('a.C.().attr', 'a'),
            ('a.C.().bttr', 'bttr'),
            ('a.b', 'a.b'),
        ]
        return s, refs

    @refs_equal
    def test_case_nested_patterns_override_but_able_to_use_simultaneously(self):
        s = 'import a\nmatch _:\n  case a.B(a, a.C(a, a.d, a), a):\n    a'
        refs = [('a', 'a'), ('a.B', 'a.B'), ('a.C', 'a.C'), ('a.d', 'a.d')]
        return s, refs

    @refs_equal
    def test_case_nested_class_attr_target_linked(self):
        s = 'import a\nmatch _:\n  case C(attr=[a.C(attr=x)]):\n    pass'
        refs = [
            ('a', 'a'), ('a.C', 'a.C'), ('a.C.().attr', 'attr'), ('a.C.().attr', 'x')
        ]
        return s, refs



"""Sphinx extension for linking code examples to reference documentation."""
import os as _os
from pathlib import Path as _Path

from sphinx.application import Sphinx
from .extension import backref, directive, SphinxCodeAutoLink
from .extension.block import clean_pycon, clean_ipython  # NOQA

_version_file = _Path(_os.path.realpath(__file__)).parent / "VERSION"
__version__ = _version_file.read_text().strip()


def setup(app: Sphinx):
    """Set up extension, directives and events."""
    state = SphinxCodeAutoLink()
    app.setup_extension('sphinx.ext.autodoc')
    app.add_css_file('sphinx-codeautolink.css')
    app.add_config_value('codeautolink_autodoc_inject', False, 'html', types=[bool])
    app.add_config_value('codeautolink_global_preface', '', 'html', types=[str])
    app.add_config_value('codeautolink_custom_blocks', {}, 'html', types=[dict])
    app.add_config_value('codeautolink_concat_default', False, 'html', types=[bool])
    app.add_config_value('codeautolink_search_css_classes', [], 'html', types=[list])
    app.add_config_value(
        'codeautolink_warn_on_missing_inventory', False, 'html', types=[bool]
    )
    app.add_config_value(
        'codeautolink_warn_on_failed_resolve', False, 'html', types=[bool]
    )

    app.add_directive('autolink-concat', directive.Concat)
    app.add_directive('autolink-examples', directive.Examples)
    app.add_directive('autolink-preface', directive.Preface)
    app.add_directive('autolink-skip', directive.Skip)

    app.connect('builder-inited', state.build_inited)
    app.connect('autodoc-process-docstring', state.autodoc_process_docstring)
    app.connect('doctree-read', state.parse_blocks)
    app.connect('env-merge-info', state.merge_environments)
    app.connect('env-purge-doc', state.purge_doc_from_environment)
    app.connect('env-updated', state.create_references)
    app.connect('doctree-resolved', state.generate_backref_tables)
    app.connect('build-finished', state.apply_links)

    app.add_node(
        backref.DetailsNode, html=(backref.visit_details, backref.depart_details)
    )
    app.add_node(
        backref.SummaryNode, html=(backref.visit_summary, backref.depart_summary)
    )
    return {'version': __version__, 'env_version': 1, 'parallel_read_safe': True}



"""Backreference tables implementation."""
from dataclasses import dataclass
from typing import Dict, List

from docutils import nodes

from .directive import DeferredExamples


@dataclass
class CodeExample:
    """Code example in the documentation."""

    document: str
    ref_id: str
    headings: List[str]


class DetailsNode(nodes.Element):
    """Collapsible details node for HTML."""

    def copy(self):
        """Copy element."""
        return self.__class__()


def visit_details(self, node: DetailsNode):
    """Insert a details tag."""
    self.body.append('<details>')


def depart_details(self, node: DetailsNode):
    """Close a details tag."""
    self.body.append('</details>')


class SummaryNode(nodes.TextElement):
    """Summary node inside a DetailsNode for HTML."""

    def copy(self):
        """Copy element."""
        return self.__class__()


def visit_summary(self, node: SummaryNode):
    """Insert a summary tag."""
    self.body.append('<summary>')


def depart_summary(self, node: SummaryNode):
    """Close a summary tag."""
    self.body.append('</summary>')


class CodeRefsVisitor(nodes.SparseNodeVisitor):
    """Replace :class:`DeferredCodeReferences` with table of concrete references."""

    def __init__(self, *args, code_refs: Dict[str, List[CodeExample]], **kwargs):
        super().__init__(*args, **kwargs)
        self.code_refs = code_refs

    def unknown_departure(self, node):
        """Ignore unknown nodes."""

    def unknown_visit(self, node):
        """Insert table in :class:`DeferredExamples`."""
        if not isinstance(node, DeferredExamples):
            return

        items = []
        for ref in self.code_refs.get(node.ref, []):
            link = ref.document + '.html'
            if ref.ref_id is not None:
                link += f'#{ref.ref_id}'
            items.append((link, ' / '.join(ref.headings)))

        items = sorted(set(items))

        if not items:
            # Remove surrounding paragraph too
            node.parent.parent.remove(node.parent)
            return

        orig_ref = node.children[0]
        node.parent.remove(node)

        # Table definition
        table = nodes.table()
        tgroup = nodes.tgroup(cols=1)
        table += tgroup
        tgroup += nodes.colspec(colwidth=1)

        if not node.collapse:
            thead = nodes.thead()
            tgroup += thead
            row = nodes.row()
            thead += row
            entry = nodes.entry()
            row += entry
            title = nodes.paragraph()
            title += nodes.Text('References to ')
            title += orig_ref
            entry += title

        tbody = nodes.tbody()
        tgroup += tbody
        for link, text in items:
            row = nodes.row()
            tbody += row

            entry = nodes.entry()
            par = nodes.paragraph()
            par += nodes.reference(internal=True, refuri=link, text=text)
            entry += par
            row += entry

        parent_par = nodes.paragraph()
        if node.collapse:
            details = DetailsNode()
            summary = SummaryNode()
            summary += nodes.Text('Expand for references to ')
            summary += orig_ref
            details += summary
            details += table
            parent_par += details
        else:
            parent_par += table
        node.parent.replace_self(parent_par)



"""Sphinx extension implementation."""
from dataclasses import dataclass
from functools import wraps
from typing import Dict, List, Optional, Set
from traceback import print_exc
from pathlib import Path

from sphinx.util import import_object
from sphinx.ext.intersphinx import InventoryAdapter

from .backref import CodeRefsVisitor, CodeExample
from .block import CodeBlockAnalyser, SourceTransform, link_html
from .directive import RemoveExtensionVisitor
from .cache import DataCache
from .resolve import resolve_location, CouldNotResolve
from ..parse import Name
from ..warn import logger, warn_type


@dataclass
class DocumentedObject:
    """Autodoc-documented code object."""

    what: str
    obj: object
    return_type: str = None


def print_exceptions(append_source: bool = False):
    """
    Print the traceback of uncaught and unexpected exceptions.

    This is done because the Sphinx process masks the traceback
    and only displays the main error message making debugging difficult.
    If append_source is set, information about the currently processed document
    is pulled from the second argument named "doctree" and added to the message.
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                print_exc()
                if append_source:
                    doctree = args[2] if len(args) > 1 else kwargs['doctree']
                    source = doctree['source']
                    msg = f'in document `{source}`'
                    if e.args:
                        e.args = (e.args[0] + f' ({msg})',) + e.args[1:]
                    else:
                        e.args = (f'Unexpected error {msg}',)
                raise
        return wrapper
    return decorator


class SphinxCodeAutoLink:
    """Provide functionality and manage state between events."""

    def __init__(self):
        # Configuration
        self.do_nothing = False
        self.global_preface: List[str] = []
        self.custom_blocks = None
        self.concat_default = None
        self.search_css_classes = None
        self.warn_missing_inventory = None
        self.warn_failed_resolve = None

        # Populated once
        self.outdated_docs: Set[str] = set()
        self.inventory = {}
        self.code_refs: Dict[str, List[CodeExample]] = {}

        # Changing state
        self.cache: Optional[DataCache] = None

    @print_exceptions()
    def build_inited(self, app):
        """Handle initial setup."""
        if app.builder.name != 'html':
            self.do_nothing = True
            return

        self.cache = DataCache(app.doctreedir, app.srcdir)
        self.cache.read()
        app.env.sphinx_codeautolink_transforms = self.cache.transforms
        self.outdated_docs = {str(Path(d)) for d in app.builder.get_outdated_docs()}
        self.custom_blocks = app.config.codeautolink_custom_blocks
        for k, v in self.custom_blocks.items():
            if isinstance(v, str):
                self.custom_blocks[k] = import_object(v)
        self.concat_default = app.config.codeautolink_concat_default
        self.search_css_classes = app.config.codeautolink_search_css_classes
        self.warn_missing_inventory = app.config.codeautolink_warn_on_missing_inventory
        self.warn_failed_resolve = app.config.codeautolink_warn_on_failed_resolve

        # Append static resources path so references in setup() are valid
        app.config.html_static_path.append(
            str(Path(__file__).parent.with_name('static').absolute())
        )

        preface = app.config.codeautolink_global_preface
        if preface:
            self.global_preface = preface.split('\n')

    @print_exceptions()
    def autodoc_process_docstring(self, app, what, name, obj, options, lines):
        """Handle autodoc-process-docstring event."""
        if self.do_nothing:
            return

        if app.config.codeautolink_autodoc_inject:
            lines.append('')
            lines.append('.. autolink-examples:: ' + name)
            lines.append('   :collapse:')

    @print_exceptions(append_source=True)
    def parse_blocks(self, app, doctree):
        """Parse code blocks for later link substitution."""
        if self.do_nothing:
            return

        visitor = CodeBlockAnalyser(
            doctree,
            source_dir=app.srcdir,
            global_preface=self.global_preface,
            custom_blocks=self.custom_blocks,
            concat_default=self.concat_default,
        )
        doctree.walkabout(visitor)
        self.cache.transforms[visitor.current_document] = visitor.source_transforms

    def merge_environments(self, app, env, docnames, other):
        """Merge transform information."""
        if self.do_nothing:
            return

        env.sphinx_codeautolink_transforms.update(
            other.sphinx_codeautolink_transforms
        )

    def purge_doc_from_environment(self, app, env, docname):
        """Remove transforms from cache."""
        if self.cache:
            self.cache.transforms.pop(docname, None)

    @staticmethod
    def make_inventory(app):
        """Create object inventory from local info and intersphinx."""
        inv_parts = {
            k: str(
                Path(app.outdir)
                / (app.builder.get_target_uri(v.docname) + f'#{v.node_id}')
            )
            for k, v in app.env.domains['py'].objects.items()
        }
        inventory = {'py:class': {
            k: (None, None, v, None) for k, v in inv_parts.items()
        }}
        inter_inv = InventoryAdapter(app.env).main_inventory
        transposed = transpose_inventory(inter_inv, relative_to=app.outdir)
        transposed.update(transpose_inventory(inventory, relative_to=app.outdir))
        return transposed

    @print_exceptions()
    def create_references(self, app, env):
        """Clean source transforms and create code references."""
        if self.do_nothing:
            return

        skipped = set()
        self.inventory = self.make_inventory(app)
        for doc, transforms in self.cache.transforms.items():
            self.filter_and_resolve(transforms, skipped, doc)
            for transform in transforms:
                for name in transform.names:
                    self.code_refs.setdefault(name.resolved_location, []).append(
                        transform.example
                    )
        if skipped and self.warn_missing_inventory:
            tops = sorted(set(s.split('.')[0] for s in skipped))
            msg = (
                f'Cannot locate modules: {str(tops)[1:-1]}'
                '\n  because of missing intersphinx or documentation entries'
            )
            logger.warning(msg, type=warn_type, subtype='missing_inventory')

    def filter_and_resolve(
        self, transforms: List[SourceTransform], skipped: Set[str], doc: str
    ):
        """Try to link name chains to objects."""
        for transform in transforms:
            filtered = []
            for name in transform.names:
                if not name.code_str:
                    continue  # empty transform target (2 calls in a row)
                try:
                    key = resolve_location(name, self.inventory)
                except CouldNotResolve:
                    if self.warn_failed_resolve:
                        path = '.'.join(name.import_components).replace('.()', '()')
                        msg = (
                            f'Could not resolve {self._resolve_msg(name)}'
                            f' using path `{path}`'
                        )
                        logger.warning(
                            msg,
                            type=warn_type,
                            subtype='failed_resolve',
                            location=(doc, transform.doc_lineno),
                        )
                    continue
                if key not in self.inventory:
                    if self.warn_missing_inventory:
                        msg = (
                            f'Inventory missing `{key}`'
                            f' when resolving {self._resolve_msg(name)}'
                        )
                        logger.warning(
                            msg,
                            type=warn_type,
                            subtype='missing_inventory',
                            location=(doc, transform.doc_lineno),
                        )
                    skipped.add(key)
                    continue
                name.resolved_location = key
                filtered.append(name)
            transform.names = filtered

    @staticmethod
    def _resolve_msg(name: Name):
        if name.lineno == name.end_lineno:
            line = f'line {name.lineno}'
        else:
            line = f'lines {name.lineno}-{name.end_lineno}'

        return f'`{name.code_str}` on {line}'

    @print_exceptions(append_source=True)
    def generate_backref_tables(self, app, doctree, docname):
        """Generate backreference tables."""
        if self.do_nothing:
            rm_vis = RemoveExtensionVisitor(doctree)
            return doctree.walkabout(rm_vis)

        visitor = CodeRefsVisitor(doctree, code_refs=self.code_refs)
        doctree.walk(visitor)

    @print_exceptions()
    def apply_links(self, app, exception):
        """Apply links to HTML output and write refs file."""
        if self.do_nothing or exception is not None:
            return

        for doc, transforms in self.cache.transforms.items():
            if not transforms or str(Path(doc)) not in self.outdated_docs:
                continue
            link_html(
                doc,
                app.outdir,
                transforms,
                self.inventory,
                self.custom_blocks,
                self.search_css_classes,
            )

        self.cache.write()


def transpose_inventory(inv: dict, relative_to: str):
    """
    Transpose Sphinx inventory from {type: {name: (..., location)}} to {name: location}.

    Also filters the inventory to Python domain only.

    Parameters
    ----------
    inv
        Sphinx inventory
    relative_to
        if a local file is found, transform it to be relative to this dir
    """
    transposed = {}
    for type_, items in inv.items():
        if not type_.startswith('py:'):
            continue
        for item, info in items.items():
            location = info[2]
            if not location.startswith('http'):
                location = str(Path(location).relative_to(relative_to))
            transposed[item] = location
    return transposed



import pytest

from ._util import refs_equal, skip_type_union, skip_walrus


class TestFunction:
    @refs_equal
    def test_func_uses(self):
        s = 'import a\ndef f():\n  a'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_async_func_uses(self):
        s = 'import a\nasync def f():\n  a'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_func_name_shadows(self):
        s = 'import a\ndef a():\n  pass\n  a'
        refs = [('a', 'a')]
        return s, refs

    @refs_equal
    def test_func_name_shadows_inside(self):
        s = 'import a\ndef a():\n  a'
        refs = [('a', 'a')]
        return s, refs

    @refs_equal
    def test_func_assigns_then_uses(self):
        s = 'import a\ndef f():\n  a = 1\n  a'
        refs = [('a', 'a')]
        return s, refs

    @refs_equal
    def test_func_assigns_then_used_outside(self):
        s = 'import a\ndef f():\n  a = 1\na'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_func_annotates_then_uses(self):
        s = 'import a\ndef f(arg: a):\n  arg.b'
        refs = [('a', 'a'), ('a', 'a'), ('a.()', 'arg'), ('a.().b', 'arg.b')]
        return s, refs

    @refs_equal
    def test_func_annotates_then_assigns(self):
        # Note: inner nodes after outer ones
        s = 'import a\ndef f(arg: a) -> a:\n  a = 1'
        refs = [('a', 'a'), ('a', 'a'), ('a', 'a'), ('a.()', 'arg')]
        return s, refs

    @refs_equal
    def test_func_annotates_as_generic_then_uses(self):
        s = 'import a\ndef f(arg: a[0]):\n  arg.b'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_func_annotates_inside_generic_then_uses(self):
        s = 'import a\ndef f(arg: b[a]):\n  arg.b'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @skip_type_union
    @refs_equal
    def test_func_annotates_union_then_uses(self):
        s = 'import a\ndef f(arg: a | 1):\n  arg.b'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_func_kw_default_uses_not_assigned(self):
        s = 'import a\ndef f(*_, c, b=a):\n  b'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_func_arg_shadows(self):
        s = 'import a\ndef f(a):\n  a'
        refs = [('a', 'a')]
        return s, refs

    @refs_equal
    def test_func_decorator_uses(self):
        s = 'import a\n@a\ndef f():\n  pass'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @pytest.mark.xfail(reason='Assignments are not tracked.')
    @refs_equal
    def test_func_uses_overrided_later(self):
        s = 'import a\ndef f():\n  a\na = 1\nf()'
        refs = [('a', 'a')]
        return s, refs

    @refs_equal
    def test_lambda_uses(self):
        s = 'import a\nlambda: a'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_lambda_arg_shadows(self):
        s = 'import a\nlambda a: a'
        refs = [('a', 'a')]
        return s, refs

    @refs_equal
    def test_lambda_arg_default_uses_not_assigned(self):
        s = 'import a\nlambda x=a: x'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_lambda_arg_default_uses_then_shadows(self):
        s = 'import a\nlambda a=a: a'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_lambda_arg_shadows_used_outside(self):
        s = 'import a\nlambda a: a\na'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_lambda_kw_default_uses(self):
        s = 'import a\nlambda *b, c, d=a: 1'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @pytest.mark.xfail(reason='No reason to do this in the real world.')
    @refs_equal
    def test_global_in_outermost_scope(self):
        s = 'import a\nglobal a'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_global_hits_import_after_inner_shadow(self):
        s = 'import a\ndef f():\n  a = 1\n  def g():\n    global a\n    a'
        refs = [('a', 'a'), ('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_global_skips_inner_import(self):
        s = 'a = 1\ndef f():\n  import a\n  def g():\n    global a\n    a'
        refs = [('a', 'a')]
        return s, refs

    @refs_equal
    def test_global_overwritten_then_used_in_inner(self):
        s = 'import a\ndef f():\n  global a\n  a = 1\n  a'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_global_overwrite_in_next_call_then_used_in_outer_before(self):
        s = 'import a\ndef f():\n  global a\n  a = 1\na\nf()'
        refs = [('a', 'a'), ('a', 'a'), ('a', 'a')]
        return s, refs

    @pytest.mark.xfail(reason='Global assigns not tracked to outer scopes.')
    @refs_equal
    def test_global_overwritten_then_used_in_outer(self):
        s = 'import a\ndef f():\n  global a\n  a = 1\nf()\na'
        refs = [('a', 'a'), ('a', 'a')]  # ref only in global statement
        return s, refs

    @refs_equal
    def test_nonlocal_hits_import_after_outer_assign(self):
        s = 'a = 1\ndef f():\n  import a\n  def g():\n    nonlocal a\n    a'
        refs = [('a', 'a'), ('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_nonlocal_skips_outer_import(self):
        s = 'import a\ndef f():\n  a = 1\n  def g():\n    nonlocal a\n    a'
        refs = [('a', 'a')]
        return s, refs

    @refs_equal
    def test_nonlocal_overwritten_then_used(self):
        s = 'a = 1\ndef f():\n  import a\n  def g():\n    nonlocal a\n    a = 1\n    a'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @pytest.mark.xfail(reason='Global deletes not tracked to outer scopes.')
    @refs_equal
    def test_global_deleted_then_used_in_outer(self):
        s = 'import a\ndef f():\n  global a\n  del a\nf()\na'
        refs = [('a', 'a'), ('a', 'a'), ('a', 'a')]  # refs in global and del
        return s, refs


class TestComprehension:
    @refs_equal
    def test_comp_uses_in_value(self):
        s = 'import a\n[a for b in c]'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_setcomp_uses_in_value(self):
        s = 'import a\n{a for b in c}'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_dictcomp_uses_in_value(self):
        s = 'import a\n{a: a for b in c}'
        refs = [('a', 'a'), ('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_generator_uses_in_value(self):
        s = 'import a\n(a for b in c)'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_comp_uses_in_ifs(self):
        s = 'import a\n[_ for _ in b if a]'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_comp_uses_in_iter(self):
        s = 'import a\n[_ for _ in b(a)]'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_comp_overrides(self):
        s = 'import a\n[a for a in b if a]'
        refs = [('a', 'a')]
        return s, refs

    @refs_equal
    def test_comp_overrides_used_after(self):
        s = 'import a\n[a for a in b]\na'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_multicomp_overrides(self):
        s = 'import a\n[a for a in b for b in a]'
        refs = [('a', 'a')]
        return s, refs

    @refs_equal
    def test_multicomp_uses_then_overrides(self):
        s = 'import a\n[a for b in c(a) for a in b]'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @skip_walrus
    @pytest.mark.xfail(reason='Assignments are not tracked outside of a "scope".')
    @refs_equal
    def test_comp_leaks_walrus(self):
        s = 'import a\n[a := i for i in b]\na'
        refs = [('a', 'a')]
        return s, refs


class TestClass:
    @refs_equal
    def test_class_name_shadows(self):
        s = 'import a\nclass a:\n  pass\na'
        refs = [('a', 'a')]
        return s, refs

    @refs_equal
    def test_class_bases_uses(self):
        s = 'import a\nclass A(a):\n  pass'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_class_keyword_uses(self):
        s = 'import a\nclass A(kw=a):\n  pass'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_class_starargs_uses(self):
        s = 'import a\nclass A(*a):\n  pass'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_class_kwargs_uses(self):
        s = 'import a\nclass A(**a):\n  pass'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_class_decorator_uses(self):
        s = 'import a\n@a\nclass A:\n  pass'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_class_body_pseudo_assigns(self):
        s = 'import a\nclass A:\n  a = 1\na'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_class_body_pseudo_shadows_for_method(self):
        s = 'import a\nclass A:\n  a = 1\n  def f(s):\n    a'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_class_method_pseudo_shadows_inside(self):
        s = 'import a\nclass A:\n  def a(s):\n    a'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_class_method_pseudo_shadows_after(self):
        s = 'import a\nclass A:\n  def a(s):\n    pass\na'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_class_method_shadows_body(self):
        s = 'import a\nclass A:\n  def a(s):\n    pass\n  a'
        refs = [('a', 'a')]
        return s, refs

    @refs_equal
    def test_class_lambda_uses_outer(self):
        s = 'import a\nclass A:\n  b = lambda: a\na = 1'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs

    @refs_equal
    def test_class_lambda_skips_body(self):
        s = 'import a\nclass A:\n  a = 2\n  b = lambda: a'
        refs = [('a', 'a'), ('a', 'a')]
        return s, refs



"""Code block processing."""
import re

from typing import List, Union, Optional, Dict, Callable, Tuple
from pathlib import Path
from dataclasses import dataclass

from bs4 import BeautifulSoup
from docutils import nodes

from ..parse import parse_names, Name, LinkContext
from .backref import CodeExample
from .directive import ConcatMarker, PrefaceMarker, SkipMarker
from ..warn import logger, warn_type

BUILTIN_BLOCKS = {
    'python': None,
    'py': None,
}


@dataclass
class SourceTransform:
    """Transforms on source code."""

    source: str
    names: List[Name]
    example: CodeExample
    doc_lineno: int


def clean_pycon(source: str) -> Tuple[str, str]:
    """Clean up Python console syntax to pure Python."""
    in_statement = False
    source = re.sub(r'^\s*<BLANKLINE>', '', source, flags=re.MULTILINE)
    clean_lines = []
    for line in source.split('\n'):
        if line.startswith('>>> '):
            in_statement = True
            clean_lines.append(line[4:])
        elif in_statement and line.startswith('... '):
            clean_lines.append(line[4:])
        else:
            in_statement = False
            clean_lines.append('')
    return source, '\n'.join(clean_lines)


BUILTIN_BLOCKS['pycon'] = clean_pycon


def clean_ipython(source: str) -> Tuple[str, str]:
    """Clean up IPython magics and console syntax to pure Python."""
    from IPython.core.inputtransformer2 import TransformerManager

    in_statement = True
    clean_lines = []
    for line in source.split('\n'):
        if re.match(r'^In \[[0-9]+\]: ', line):
            in_statement = True
        elif re.match(r'^Out\[[0-9]+\]: ', line):
            in_statement = False
        clean_lines.append(line * in_statement)

    return source, TransformerManager().transform_cell('\n'.join(clean_lines))


try:
    import IPython
except ImportError:
    pass
else:
    del IPython
    BUILTIN_BLOCKS['ipython'] = clean_ipython
    BUILTIN_BLOCKS['ipython3'] = clean_ipython


class CodeBlockAnalyser(nodes.SparseNodeVisitor):
    """Transform literal blocks of Python with links to reference documentation."""

    def __init__(
        self,
        *args,
        source_dir: str,
        global_preface: List[str],
        custom_blocks: Dict[str, Callable[[str], str]],
        concat_default: bool,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.source_transforms: List[SourceTransform] = []
        relative_path = Path(self.document['source']).relative_to(source_dir)
        self.current_document = str(relative_path.with_suffix(''))
        self.global_preface = global_preface
        self.transformers = BUILTIN_BLOCKS.copy()
        self.transformers.update(custom_blocks)
        self.valid_blocks = self.transformers.keys()
        self.title_stack = []
        self.current_refid = None
        self.prefaces = []
        self.concat_global = concat_default
        self.concat_section = False
        self.concat_sources = []
        self.skip = None

    def unknown_visit(self, node):
        """Handle and delete custom directives, ignore others."""
        if isinstance(node, ConcatMarker):
            if node.mode not in ('off', 'section', 'on'):
                msg = f'Invalid concatenation argument: `{node.mode}`'
                logger.error(
                    msg, type=warn_type, subtype='invalid_argument', location=node
                )

            self.concat_sources = []
            if node.mode == 'section':
                self.concat_section = True
            else:
                self.concat_section = False
                self.concat_global = (node.mode == 'on')
            node.parent.remove(node)
        elif isinstance(node, PrefaceMarker):
            self.prefaces.extend(node.content.split('\n'))
            node.parent.remove(node)
        elif isinstance(node, SkipMarker):
            if node.level not in ('next', 'section', 'file', 'off'):
                msg = f'Invalid skipping argument: `{node.level}`'
                logger.error(
                    msg, type=warn_type, subtype='invalid_argument', location=node
                )
            self.skip = node.level if node.level != 'off' else None
            node.parent.remove(node)

    def unknown_departure(self, node):
        """Ignore unknown nodes."""

    def visit_title(self, node):
        """Track section names and break concatenation and skipping."""
        self.title_stack.append(node.astext())
        if self.concat_section:
            self.concat_section = False
            self.concat_sources = []
        if self.skip == 'section':
            self.skip = None

    def visit_section(self, node):
        """Record first section ID."""
        self.current_refid = node['ids'][0]

    def depart_section(self, node):
        """Pop latest title."""
        self.title_stack.pop()

    def visit_doctest_block(self, node):
        """Visit a Python doctest block."""
        return self.parse_source(node, 'pycon')

    def visit_literal_block(self, node: nodes.literal_block):
        """Visit a generic literal block."""
        return self.parse_source(node, node.get('language', None))

    def parse_source(
        self,
        node: Union[nodes.literal_block, nodes.doctest_block],
        language: Optional[str]
    ):
        """Analyse Python code blocks."""
        prefaces = self.prefaces
        self.prefaces = []

        skip = self.skip
        if skip == 'next':
            self.skip = None

        if (
            skip
            or len(node.children) != 1
            or not isinstance(node.children[0], nodes.Text)
            or language not in self.valid_blocks
        ):
            return

        source = node.children[0].astext()
        transformer = self.transformers[language]
        if transformer:
            try:
                source, clean_source = transformer(source)
            except SyntaxError as e:
                show_source = self._format_source_for_error(source, prefaces)
                msg = self._parsing_error_msg(e, language, show_source)
                logger.warning(
                    msg, type=warn_type, subtype='parse_block', location=node
                )
                return
        else:
            clean_source = source
        example = CodeExample(
            self.current_document, self.current_refid, list(self.title_stack)
        )
        transform = SourceTransform(source, [], example, node.line)
        self.source_transforms.append(transform)

        modified_source = '\n'.join(
            self.global_preface
            + self.concat_sources
            + prefaces
            + [clean_source]
        )
        try:
            names = parse_names(modified_source, node)
        except SyntaxError as e:
            show_source = self._format_source_for_error(source, prefaces)
            msg = self._parsing_error_msg(e, language, show_source)
            logger.warning(msg, type=warn_type, subtype='parse_block', location=node)
            return

        if prefaces or self.concat_sources or self.global_preface:
            concat_lens = [s.count('\n') + 1 for s in self.concat_sources]
            hidden_len = (
                len(prefaces) + sum(concat_lens) + len(self.global_preface)
            )
            for name in names:
                name.lineno -= hidden_len
                name.end_lineno -= hidden_len

        if self.concat_section or self.concat_global:
            self.concat_sources.extend(prefaces + [clean_source])

        # Remove transforms from concatenated sources
        transform.names.extend([n for n in names if n.lineno > 0])

    def _format_source_for_error(self, source: str, prefaces: List[str]) -> str:
        split_source = source.split('\n')
        guides = [''] * len(split_source)
        ix = 0
        if self.global_preface:
            guides[0] = 'global preface:'
            ix += len(self.global_preface)
        if self.concat_sources:
            guides[ix] = 'concatenations:'
            ix += len(self.concat_sources)
        if prefaces:
            guides[ix] = 'local preface:'
            ix += len(prefaces)
        guides[ix] = 'block source:'
        pad = max(len(i) + 1 for i in guides)
        guides = [g.ljust(pad) for g in guides]
        return '\n'.join([g + s for g, s in zip(guides, split_source)])

    def _parsing_error_msg(self, error: Exception, language: str, source: str) -> str:
        return '\n'.join([
            str(error) + f' in document "{self.current_document}"',
            f'Parsed source in `{language}` block:',
            source,
        ])


def link_html(
    document: str,
    out_dir: str,
    transforms: List[SourceTransform],
    inventory: dict,
    custom_blocks: dict,
    search_css_classes: list,
):
    """Inject links to code blocks on disk."""
    html_file = Path(out_dir) / (document + '.html')
    text = html_file.read_text('utf-8')
    soup = BeautifulSoup(text, 'html.parser')

    block_types = BUILTIN_BLOCKS.keys() | custom_blocks.keys()
    classes = [f'highlight-{t}' for t in block_types] + ['doctest']
    classes += search_css_classes

    blocks = []
    for c in classes:
        blocks.extend(list(soup.find_all('div', attrs={'class': c})))
    unique_blocks = {b.sourceline: b for b in blocks}.values()
    blocks = sorted(unique_blocks, key=lambda b: b.sourceline)
    inners = [block.select('div > pre')[0] for block in blocks]

    up_lvls = len(html_file.relative_to(out_dir).parents) - 1
    local_prefix = '../' * up_lvls
    link_pattern = (
        '<a href="{link}" title="{title}" class="sphinx-codeautolink-a">{text}</a>'
    )

    for trans in transforms:
        for ix in range(len(inners)):
            if trans.source.rstrip() == ''.join(inners[ix].strings).rstrip():
                inner = inners.pop(ix)
                break
        else:
            msg = f'Could not match a code example to HTML, source:\n{trans.source}'
            logger.warning(
                msg, type=warn_type, subtype='match_block', location=document
            )
            continue

        lines = str(inner).split('\n')

        for name in trans.names:
            begin_line = name.lineno - 1
            end_line = name.end_lineno - 1
            selection = '\n'.join(lines[begin_line:end_line + 1])

            # Reverse because a.b = a.b should replace from the right
            matches = list(re.finditer(construct_name_pattern(name), selection))[::-1]
            if not matches:
                msg = (
                    f'Could not match transformation of `{name.code_str}` '
                    f'on source lines {name.lineno}-{name.end_lineno}, '
                    f'source:\n{trans.source}'
                )
                logger.warning(
                    msg, type=warn_type, subtype='match_name', location=document
                )
                continue

            start, end = matches[0].span()
            start += len(matches[0].group(1))
            location = inventory[name.resolved_location]
            if not any(location.startswith(s) for s in ('http://', 'https://')):
                location = local_prefix + location
            link = link_pattern.format(
                link=location,
                title=name.resolved_location,
                text=selection[start:end]
            )
            transformed = selection[:start] + link + selection[end:]
            lines[begin_line:end_line + 1] = transformed.split('\n')

        inner.replace_with(BeautifulSoup('\n'.join(lines), 'html.parser'))

    html_file.write_text(str(soup), 'utf-8')


# ---------------------------------------------------------------
# Patterns for different types of name access in highlighted HTML
# ---------------------------------------------------------------
period = r'\s*<span class="o">.</span>\s*'
name_pattern = '<span class="n">{name}</span>'
# Pygments has special classes for different types of nouns
# which are also highlighted in import statements
first_name_pattern = '<span class="[a-z]+">@?{name}</span>'
import_target_pattern = '<span class="[a-z]+">{name}</span>'
import_from_pattern = '<span class="nn">{name}</span>'

# The builtin re doesn't support variable-width lookbehind,
# so instead we use a match groups in all pre patterns to remove the non-content.
no_dot_prere = r'(?<!<span class="o">\.</span>)()'
# Potentially instead assert an initial closing parenthesis followed by a dot.
call_dot_prere = r'(\)</span>\s*<span class="o">\.</span>\s*)'
import_prere = (
    r'((<span class="kn">import</span>\s+(<span class="p">\(</span>\s*)?)'
    r'|(<span class="p">,</span>\s*))'
)
from_prere = r'(<span class="kn">from</span>\s+)'

no_dot_postre = r'(?!(<span class="o">\.)|(</a>))'
import_postre = r'(?=($)|(\s+)|(<span class="p">,</span>)|(<span class="p">\)))(?!</a>)'
from_postre = r'(?=\s*<span class="kn">import</span>)'


def construct_name_pattern(name: Name) -> str:
    """Construct a regex pattern for searching a name in HTML."""
    if name.context == LinkContext.none:
        parts = name.code_str.split('.')
        pattern = period.join(
            [first_name_pattern.format(name=parts[0])]
            + [name_pattern.format(name=p) for p in parts[1:]]
        )
        return no_dot_prere + pattern + no_dot_postre
    elif name.context == LinkContext.after_call:
        parts = name.code_str.split('.')
        pattern = period.join(
            [first_name_pattern.format(name=parts[0])]
            + [name_pattern.format(name=p) for p in parts[1:]]
        )
        return call_dot_prere + pattern + no_dot_postre
    elif name.context == LinkContext.import_from:
        pattern = import_from_pattern.format(name=name.code_str)
        return from_prere + pattern + from_postre
    elif name.context == LinkContext.import_target:
        pattern = import_target_pattern.format(name=name.code_str)
        return import_prere + pattern + import_postre



"""Analyse AST of code blocks to determine used names and their sources."""
import ast
import sys
import builtins

from contextlib import contextmanager
from enum import Enum
from functools import wraps
from importlib import import_module
from typing import Dict, Union, List, Optional, Tuple
from dataclasses import dataclass, field

from .warn import logger, warn_type

HAS_WALRUS = (sys.version_info >= (3, 8))
HAS_MATCH = (sys.version_info >= (3, 10))


def parse_names(source: str, doctree_node) -> List['Name']:
    """Parse names from source."""
    tree = ast.parse(source)
    visitor = ImportTrackerVisitor(doctree_node)
    visitor.visit(tree)
    return visitor.accessed


def linenos(node: ast.AST) -> Tuple[int, int]:
    """Return lineno and end_lineno safely."""
    return node.lineno, getattr(node, 'end_lineno', node.lineno)


@dataclass
class Component:
    """Name access component."""

    name: str
    lineno: int
    end_lineno: int
    context: str  # as in ast.Load / Store / Del

    @classmethod
    def from_ast(cls, node):
        """Generate a Component from an AST node."""
        context = 'load'
        if isinstance(node, ast.Name):
            name = node.id
            context = node.ctx.__class__.__name__.lower()
        elif isinstance(node, ast.Attribute):
            name = node.attr
            context = node.ctx.__class__.__name__.lower()
        elif isinstance(node, ast.arg):
            name = node.arg
        elif isinstance(node, ast.Call):
            name = NameBreak.call
        elif HAS_MATCH and isinstance(node, ast.MatchAs):
            name = node.name
            context = 'store'
        else:
            raise ValueError(f'Invalid AST for component: {node.__class__.__name__}')
        return cls(name, *linenos(node), context)


@dataclass
class PendingAccess:
    """Pending name access."""

    components: List[Component]


@dataclass
class AssignTarget:
    """
    Assign target.

    `elements` represent the flattened assignment target elements.
    If a single PendingAccess is found, it should be used to store the value
    on the right hand side of the assignment. If multiple values are found,
    the assignment contained unpacking, and only overwriting of names should occur.
    """

    elements: List[Optional[PendingAccess]]


@dataclass
class Assignment:
    """
    Representation of an assignment statement.

    - ordinarily one value to a single target
    - multiple targets when chain assigning (a = b = c)
    - nested assignments in walruses (a = b := c)
    """

    targets: List[AssignTarget]
    value: Union[PendingAccess, 'Assignment', None]


class NameBreak(str, Enum):
    """Elements that break name access chains."""

    call = '()'


class LinkContext(str, Enum):
    """Context in which a link appears."""

    none = 'none'
    after_call = 'after_call'
    import_from = 'import_from'  # from *mod.sub* import foo
    import_target = 'import_target'  # from mod.sub import *foo*


@dataclass
class Name:
    """A name accessed in the source traced back to an import."""

    import_components: List[str]
    code_str: str
    lineno: int
    end_lineno: int
    context: Optional[LinkContext] = None
    resolved_location: Optional[str] = None


@dataclass
class Access:
    """
    Accessed import, to be broken down into suitable chunks.

    :attr:`prior_components` are components that are implicitly used via
    the base name in :attr:`components`, which is the part that shows on the line.
    :attr:`hidden_components` is an attribute of split Access, in which the
    proper components are not moved to prior components to track which were
    present on the line of the access.

    The base component that connects an import to the name that was used to
    access it is automatically removed from the components in :attr:`full_components`.
    """

    context: LinkContext
    prior_components: List[Component]
    components: List[Component]
    hidden_components: List[Component] = field(default_factory=list)

    @property
    def full_components(self):
        """All components from import base to used components."""
        if not self.prior_components:
            # Import statement itself
            return self.hidden_components + self.components

        if self.hidden_components:
            proper_components = self.hidden_components[1:] + self.components
        else:
            proper_components = self.components[1:]
        return self.prior_components + proper_components

    @property
    def code_str(self):
        """Code representation of components."""
        break_on = set(NameBreak)
        breaks = [i for i, c in enumerate(self.components) if c.name in break_on]
        start_ix = breaks[-1] + 1 if breaks else 0
        return '.'.join(c.name for c in self.components[start_ix:])

    @property
    def lineno_span(self) -> Tuple[int, int]:
        """Estimate the lineno span of components."""
        min_ = min(c.lineno for c in self.components)
        max_ = max(c.end_lineno for c in self.components)
        return min_, max_

    @staticmethod
    def to_name(instance: 'Access') -> Name:
        """Convert access to name."""
        return Name(
            [c.name for c in instance.full_components],
            instance.code_str,
            *instance.lineno_span,
            context=instance.context,
        )

    def split(self) -> List[Name]:
        """Split access into multiple names."""
        # Copy to avoid modifying the instance in place
        items = [Access(
            context=self.context,
            prior_components=self.prior_components[:],
            components=self.components[:],
            hidden_components=self.hidden_components[:],
        )]
        while True:
            current = items[-1]
            for i, comp in enumerate(current.components):
                if i and comp.name == NameBreak.call:
                    hidden = current.hidden_components + current.components[:i]
                    next_ = Access(
                        LinkContext.after_call,
                        current.prior_components,
                        current.components[i:],
                        hidden_components=hidden,
                    )
                    current.components = current.components[:i]
                    items.append(next_)
                    break
            else:
                break
        if items[-1].components[-1].name == NameBreak.call:
            items.pop()
        return [self.to_name(i) for i in items]


def track_parents(func):
    """
    Track a stack of nodes to determine the position of the current node.

    Uses and increments the surrounding classes :attr:`_parents`.
    """
    @wraps(func)
    def wrapper(self: 'ImportTrackerVisitor', *args, **kwargs):
        self._parents += 1
        result = func(self, *args, **kwargs)
        self._parents -= 1
        if not self._parents:
            self.dispatch_result(result)
        return result
    return wrapper


builtin_components: Dict[str, List[Component]] = {
    b: [Component(b, -1, -1, LinkContext.none)] for b in dir(builtins)
}


class ImportTrackerVisitor(ast.NodeVisitor):
    """Track imports and their use through source code."""

    def __init__(self, doctree_node):
        super().__init__()
        self.accessed: List[Name] = []
        self.in_augassign = False
        self._parents = 0
        self._no_split = False
        self.doctree_node = doctree_node

        # Stack for dealing with class body pseudo scopes
        # which are completely bypassed by inner scopes (func, lambda).
        # Current values are copied to the next class body level.
        self.pseudo_scopes_stack: List[Dict[str, List[Component]]] = [
            builtin_components.copy()
        ]
        # Stack for dealing with nested scopes.
        # Holds references to the values of previous nesting levels.
        self.outer_scopes_stack: List[Dict[str, List[Component]]] = []

    def save_access(self, access: Access) -> None:
        """Convert Access to Names to store in the visitor for aggregation."""
        names = access.split() if not self._no_split else [Access.to_name(access)]
        self.accessed.extend(names)

    @contextmanager
    def no_split(self):
        """Disable splitting Accesses."""
        self._no_split, old = (True, self._no_split)
        yield
        self._no_split = old

    @contextmanager
    def reset_parents(self):
        """Reset parents state for the duration of the context."""
        self._parents, old = (0, self._parents)
        yield
        self._parents = old

    # Nodes that are excempt from resetting parents in default visit
    track_nodes = (
        ast.Name,
        ast.Attribute,
        ast.Call,
    )
    if HAS_WALRUS:
        track_nodes += (ast.NamedExpr,)
    if HAS_MATCH:
        track_nodes += (ast.MatchAs,)

    def visit(self, node: ast.AST):
        """Override default visit to track name access and assignments."""
        if isinstance(node, self.track_nodes):
            return super().visit(node)

        with self.reset_parents():
            return super().visit(node)

    def overwrite_name(self, name: str):
        """Overwrite name in current scope."""
        # Technically dotted values could now be bricked,
        # but we can't prevent the earlier values in the chain from being used.
        # There is a chance that the value which was assigned is a something
        # that we could follow, but for now it's not really worth the effort.
        # With a dotted value, the following condition will never hold as long
        # as the dotted components of imports are discarded on creating the import.
        self.pseudo_scopes_stack[-1].pop(name, None)

    def assign_name(self, name: str, components: List[Component]):
        """Import or assign a name to current scope."""
        # Overwriting technically unnecessary until it properly follows dots
        self.overwrite_name(name)
        self.pseudo_scopes_stack[-1][name] = components

    def create_access(
        self, scope_key: str, new_components: List[Component]
    ) -> Optional[Access]:
        """Create access from scope."""
        prior = self.pseudo_scopes_stack[-1].get(scope_key, None)
        if prior is None:
            return

        access = Access(LinkContext.none, prior, new_components)
        self.save_access(access)
        return access

    def resolve_pending_access(self, pending: PendingAccess) -> Optional[Access]:
        """Resolve and save pending access."""
        components = pending.components

        context = components[0].context
        if context == 'store' and not self.in_augassign:
            self.overwrite_name(components[0].name)
            return

        access = self.create_access(components[0].name, components)
        if context == 'del':
            self.overwrite_name(components[0].name)
        return access

    def resolve_assignment(self, assignment: Assignment) -> Optional[Access]:
        """Resolve access for assignment values and targets."""
        access = self.dispatch_result(assignment.value)
        self._resolve_assign_targets(assignment, access)
        return access

    def _resolve_assign_targets(self, assignment: Assignment, access: Access):
        for assign in assignment.targets:
            if assign is None:
                continue

            # On multiple nested targets, only overwrite assigned names
            value = access if len(assign.elements) <= 1 else None

            for target in assign.elements:
                self._resolve_assign_target(target, value)

    def _resolve_assign_target(
        self, target: Optional[PendingAccess], value: Optional[Access]
    ):
        if target is None:
            return

        if len(target.components) == 1:
            comp = target.components[0]
            if value is None:
                self.overwrite_name(comp.name)
            else:
                self.assign_name(comp.name, value.full_components)
                self.create_access(comp.name, target.components)
        else:
            self.resolve_pending_access(target)

    def create_simple_access(self, name: str, lineno: int) -> None:
        """Create single-component access to scope."""
        component = Component(name, lineno, lineno, 'load')
        self.create_access(component.name, [component])

    def dispatch_result(
        self, result: Union[PendingAccess, Assignment, None]
    ) -> Optional[Access]:
        """Determine the appropriate processing after tracking an access chain."""
        if isinstance(result, Assignment):
            return self.resolve_assignment(result)
        elif isinstance(result, PendingAccess):
            return self.resolve_pending_access(result)

    def visit_Global(self, node: ast.Global):
        """Import from top scope."""
        if not self.outer_scopes_stack:
            return  # in outermost scope already, no-op for imports

        imports = self.outer_scopes_stack[0]
        for name in node.names:
            self.overwrite_name(name)
            if name in imports:
                self.assign_name(name, imports[name])
                self.create_simple_access(name, node.lineno)

    def visit_Nonlocal(self, node: ast.Nonlocal):
        """Import from intermediate scopes."""
        imports_stack = self.outer_scopes_stack[1:]
        for name in node.names:
            self.overwrite_name(name)
            for imports in imports_stack[::-1]:
                if name in imports:
                    self.assign_name(name, imports[name])
                    self.create_simple_access(name, node.lineno)
                    break

    def visit_Import(self, node: Union[ast.Import, ast.ImportFrom], prefix: str = ''):
        """Register import source."""
        import_star = (node.names[0].name == '*')
        if import_star:
            try:
                mod = import_module(node.module)
                import_names = [
                    name for name in mod.__dict__ if not name.startswith('_')
                ]
                aliases = [None] * len(import_names)
            except ImportError:
                logger.warning(
                    f'Could not import module `{node.module}` for parsing!',
                    type=warn_type,
                    subtype='import_star',
                    location=self.doctree_node,
                )
                import_names = []
                aliases = []
        else:
            import_names = [name.name for name in node.names]
            aliases = [name.asname for name in node.names]

        prefix_parts = prefix.rstrip('.').split('.') if prefix else []
        prefix_components = [
            Component(n, *linenos(node), 'load') for n in prefix_parts
        ]
        if prefix:
            self.save_access(Access(LinkContext.import_from, [], prefix_components))

        for import_name, alias in zip(import_names, aliases):
            if not import_star:
                components = [
                    Component(n, *linenos(node), 'load')
                    for n in import_name.split('.')
                ]
                self.save_access(
                    Access(LinkContext.import_target, [], components, prefix_components)
                )

            if not alias and '.' in import_name:
                # equivalent to only import top level module since we don't
                # follow assignments and the outer modules also get imported
                import_name = import_name.split('.')[0]

            full_components = [
                Component(n, *linenos(node), 'store')
                for n in (prefix + import_name).split('.')
            ]
            self.assign_name(alias or import_name, full_components)

    def visit_ImportFrom(self, node: ast.ImportFrom):
        """Register import source."""
        if node.level:  # relative import
            for name in node.names:
                self.overwrite_name(name.asname or name.name)
        else:
            self.visit_Import(node, prefix=node.module + '.')

    @track_parents
    def visit_Name(self, node: ast.Name):
        """Visit a Name node."""
        return PendingAccess([Component.from_ast(node)])

    @track_parents
    def visit_Attribute(self, node: ast.Attribute):
        """Visit an Attribute node."""
        inner: Optional[PendingAccess] = self.visit(node.value)
        if inner is not None:
            inner.components.append(Component.from_ast(node))
        return inner

    @track_parents
    def visit_Call(self, node: ast.Call):
        """Visit a Call node."""
        inner: Optional[PendingAccess] = self.visit(node.func)
        if inner is not None:
            inner.components.append(Component.from_ast(node))
        with self.reset_parents():
            for arg in node.args + node.keywords:
                self.visit(arg)
            if hasattr(node, 'starargs'):
                self.visit(node.starargs)
            if hasattr(node, 'kwargs'):
                self.visit(node.kwargs)
        return inner

    @track_parents
    def visit_Tuple(self, node: ast.Tuple):
        """Visit a Tuple node."""
        if isinstance(node.ctx, ast.Store):
            accesses = []
            for element in node.elts:
                ret = self.visit(element)
                if isinstance(ret, PendingAccess) or ret is None:
                    accesses.append(ret)
                else:
                    accesses.extend(ret)
            return accesses
        else:
            with self.reset_parents():
                for element in node.elts:
                    self.visit(element)

    @track_parents
    def visit_Assign(self, node: ast.Assign):
        """Visit an Assign node."""
        value = self.visit(node.value)
        targets = []
        for n in node.targets[::-1]:
            target = self.visit(n)
            if not isinstance(target, list):
                target = [target]
            targets.append(AssignTarget(target))
        return Assignment(targets, value)

    @track_parents
    def visit_AnnAssign(self, node: ast.AnnAssign):
        """Visit an AnnAssign node."""
        value = self.visit(node.value) if node.value is not None else None
        annot = self.visit(node.annotation)
        if annot is not None:
            if value is not None:
                self.resolve_pending_access(value)

            annot.components.append(
                Component(NameBreak.call, *linenos(node.annotation), 'load')
            )
            value = annot

        target = self.visit(node.target)
        if value is not None:
            return Assignment([AssignTarget([target])], value)

    def visit_AugAssign(self, node: ast.AugAssign):
        """Visit an AugAssign node."""
        self.visit(node.value)
        self.in_augassign, temp = (True, self.in_augassign)
        self.visit(node.target)
        self.in_augassign = temp

    @track_parents
    def visit_NamedExpr(self, node):
        """Visit a NamedExpr node."""
        value = self.visit(node.value)
        target = self.visit(node.target)
        return Assignment([AssignTarget([target])], value)

    @track_parents
    def visit_MatchClass(self, node):
        """Visit a match case class as a series of assignments."""
        with self.reset_parents():
            cls = self.visit(node.cls)

        accesses = []
        for n in node.patterns:
            access = self.visit(n)
            if access is not None:
                accesses.append(access)

        assigns = []
        for attr, pattern in zip(node.kwd_attrs, node.kwd_patterns):
            target = self.visit(pattern)
            attr_comps = [
                Component(NameBreak.call, *linenos(node), 'load'),
                Component(attr, *linenos(node), 'load'),
            ]
            access = PendingAccess(cls.components + attr_comps)
            assigns.append(Assignment([AssignTarget([target])], access))

        for access in accesses:
            self.resolve_pending_access(access)

        with self.no_split():
            for assign in assigns:
                self.resolve_assignment(assign)

    @track_parents
    def visit_MatchAs(self, node):
        """Track match alias names."""
        return PendingAccess([Component.from_ast(node)])

    def visit_AsyncFor(self, node: ast.AsyncFor):
        """Delegate to sync for."""
        self.visit_For(node)

    def visit_For(self, node: Union[ast.For, ast.AsyncFor]):
        """Swap node order."""
        self.visit(node.iter)
        self.visit(node.target)
        for n in node.body:
            self.visit(n)
        for n in node.orelse:
            self.visit(n)

    def visit_ClassDef(self, node: ast.ClassDef):
        """Handle pseudo scope of class body."""
        for dec in node.decorator_list:
            self.visit(dec)
        for base in node.bases:
            self.visit(base)
        for kw in node.keywords:
            self.visit(kw)

        self.overwrite_name(node.name)
        self.pseudo_scopes_stack.append(self.pseudo_scopes_stack[0].copy())
        for b in node.body:
            self.visit(b)
        self.pseudo_scopes_stack.pop()

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef):
        """Delegate to func def."""
        self.visit_FunctionDef(node)

    @staticmethod
    def _get_args(node: ast.arguments):
        posonly = getattr(node, 'posonlyargs', [])  # only on 3.8+
        return node.args + node.kwonlyargs + posonly

    def visit_FunctionDef(self, node: Union[ast.FunctionDef, ast.AsyncFunctionDef]):
        """Swap node order and separate inner scope."""
        self.overwrite_name(node.name)
        for dec in node.decorator_list:
            self.visit(dec)
        for d in node.args.defaults + node.args.kw_defaults:
            if d is None:
                continue
            self.visit(d)
        args = self._get_args(node.args)
        args += [node.args.vararg, node.args.kwarg]

        inner = self.__class__(self.doctree_node)
        inner.pseudo_scopes_stack[0] = self.pseudo_scopes_stack[0].copy()
        inner.outer_scopes_stack = list(self.outer_scopes_stack)
        inner.outer_scopes_stack.append(self.pseudo_scopes_stack[0])

        for arg in args:
            if arg is None:
                continue
            inner.visit(arg)
        if node.returns is not None:
            self.visit(node.returns)
        for n in node.body:
            inner.visit(n)
        self.accessed.extend(inner.accessed)

    @track_parents
    def visit_arg(self, arg: ast.arg):
        """Handle function argument and its annotation."""
        target = PendingAccess([Component.from_ast(arg)])
        if arg.annotation is not None:
            value = self.visit(arg.annotation)
            if value is not None:
                value.components.append(
                    Component(NameBreak.call, *linenos(arg), 'load')
                )
        else:
            value = None
        return Assignment([AssignTarget([target])], value)

    def visit_Lambda(self, node: ast.Lambda):
        """Swap node order and separate inner scope."""
        for d in node.args.defaults + node.args.kw_defaults:
            if d is None:
                continue
            self.visit(d)
        args = self._get_args(node.args)
        args += [node.args.vararg, node.args.kwarg]

        inner = self.__class__(self.doctree_node)
        inner.pseudo_scopes_stack[0] = self.pseudo_scopes_stack[0].copy()
        for arg in args:
            if arg is None:
                continue
            inner.overwrite_name(arg.arg)
        inner.visit(node.body)
        self.accessed.extend(inner.accessed)

    def visit_ListComp(self, node: ast.ListComp):
        """Delegate to generic comp."""
        self.visit_generic_comp([node.elt], node.generators)

    def visit_SetComp(self, node: ast.SetComp):
        """Delegate to generic comp."""
        self.visit_generic_comp([node.elt], node.generators)

    def visit_DictComp(self, node: ast.DictComp):
        """Delegate to generic comp."""
        self.visit_generic_comp([node.key, node.value], node.generators)

    def visit_GeneratorExp(self, node: ast.GeneratorExp):
        """Delegate to generic comp."""
        self.visit_generic_comp([node.elt], node.generators)

    def visit_comprehension(self, node: ast.comprehension):
        """Swap node order."""
        self.visit(node.iter)
        self.visit(node.target)
        for f in node.ifs:
            self.visit(f)

    def visit_generic_comp(
        self, values: List[ast.AST], generators: List[ast.comprehension]
    ):
        """Separate inner scope, respects class body scope."""
        inner = self.__class__(self.doctree_node)
        inner.pseudo_scopes_stack[0] = self.pseudo_scopes_stack[-1].copy()
        for gen in generators:
            inner.visit(gen)
        for value in values:
            inner.visit(value)
        self.accessed.extend(inner.accessed)



import sys
import pytest

from functools import wraps
from sphinx_codeautolink.parse import parse_names

skip_walrus = pytest.mark.skipif(
    sys.version_info < (3, 8),
    reason='Walrus introduced in Python 3.8.',
)

skip_type_union = pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason='Type union introduced in Python 3.10.',
)

skip_match = pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason='Match introduced in Python 3.10.',
)


def refs_equal(func):
    @wraps(func)
    def wrapper(self):
        source, expected = func(self)
        names = parse_names(source, doctree_node=None)
        names = sorted(names, key=lambda name: name.lineno)
        print('Source:\n' + source)
        print('\nExpected names:')
        for components, code_str in expected:
            print(f'components={components}, code_str={code_str}')
        print('\nParsed names:')
        [print(n) for n in names]
        for n, e in zip(names, expected):
            s = '.'.join(c for c in n.import_components)
            assert s == e[0], f'Wrong import! Expected\n{e}\ngot\n{n}'
            assert n.code_str == e[1], f'Wrong code str! Expected\n{e}\ngot\n{n}'

        msg = f'Wrong number of nodes! Expected {len(expected)}, got {len(names)}'
        assert len(names) == len(expected), msg
    return wrapper



import requests
from pathlib import Path
from bs4 import BeautifulSoup

# Cache external pages for the duration of the runtime,
# so that we don't request them multiple times needlessly
sess = requests.Session()
external_site_ids = {}


def check_link_targets(root: Path) -> int:
    """Validate links in HTML site at root, return number of links found."""
    site_docs = {
        p.relative_to(root): BeautifulSoup(p.read_text('utf-8'), 'html.parser')
        for p in root.glob('**/*.html')
    }
    site_ids = {k: gather_ids(v) for k, v in site_docs.items()}

    total = 0
    for doc, soup in site_docs.items():
        doc = str(doc)
        for link in soup.find_all('a', attrs={'class': 'sphinx-codeautolink-a'}):
            base, id_ = link['href'].split('#')
            if any(base.startswith(s) for s in ('http://', 'https://')):
                if base not in external_site_ids:
                    soup = BeautifulSoup(sess.get(base).text, 'html.parser')
                    external_site_ids[base] = gather_ids(soup)
                ids = external_site_ids[base]
            else:
                ids = site_ids[Path(base)]
            assert id_ in ids, (
                f'ID {id_} not found in {base}'
                f' while validating link for `{link.string}` in {doc}!'
            )
            total += 1
    return total


def gather_ids(soup: BeautifulSoup) -> set:
    """Gather all HTML IDs from a given page."""
    return set(tag['id'] for tag in soup.find_all(id=True))



"""Docstring."""
from typing import Optional, Union

from .sub import subfoo  # NOQA


class Baz:
    """Baz test class."""

    bute = 1


class Foo:
    """Foo test class."""

    attr: str = 'test'
    type_attr = Baz

    def meth(self) -> Baz:
        """Test method."""

    def selfref(self) -> "Foo":
        """Return self."""

    def __call__(self) -> Baz:
        """Test call."""


def bar() -> Foo:
    """bar test function."""


def optional() -> Optional[Foo]:
    """Return optional type."""


def optional_manual() -> Union[None, Foo]:
    """Return manually constructed optional type."""


def optional_counter() -> Union[Foo, Baz]:
    """Failing case for incorrect optional type handling."""


def compile():
    """Shadows built in compile function."""


class Child(Foo):
    """Foo child class."""



"""Extension data cache."""
import json

from dataclasses import asdict
from typing import Dict, List
from pathlib import Path

from .block import SourceTransform, CodeExample, Name


class DataCache:
    """Data cache."""

    cache_filename = 'codeautolink-cache.json'

    def __init__(self, cache_dir: str, src_dir: str):
        self.cache_dir: Path = Path(cache_dir)
        self.src_dir: Path = Path(src_dir)
        self.transforms: Dict[str, List[SourceTransform]] = {}

    def read(self):
        """Read from cache."""
        cache = self.cache_dir / self.cache_filename
        if not cache.exists():
            return
        content = json.loads(cache.read_text('utf-8'))
        for file, transforms in content.items():
            full_path = self.src_dir / (file + '.rst')
            if not full_path.exists():
                continue
            for transform in transforms:
                transform['example'] = CodeExample(**transform['example'])
                transform['names'] = [Name(**n) for n in transform['names']]
            self.transforms[file] = [SourceTransform(**t) for t in transforms]

    def write(self):
        """Write to cache."""
        cache = self.cache_dir / self.cache_filename
        transforms_dict = {}
        for file, transforms in self.transforms.items():
            transforms_dict[file] = [asdict(t) for t in transforms]
        cache.write_text(json.dumps(transforms_dict, indent=2), 'utf-8')



from typing import List


class Knight:
    limbs: int = 4
    taunts: List[str] = [
        "None shall pass!",
        "'Tis but a scratch!",
        "It's just a flesh wound... Chicken!",
        "Right, I'll do you for that!",
        "Oh, I see, running away?",
    ]

    def scratch(self) -> None:
        """Scratch the knight."""
        self.limbs -= 1

    def taunt(self) -> str:
        """Knight taunts the adversary."""
        return self.taunts[::-1][self.limbs]


class Shrubbery:
    """A shrubbery bought in town."""

    looks_nice: bool
    too_expensive: bool

    def __init__(self, looks_nice: bool, too_expensive: bool):
        self.looks_nice = looks_nice
        self.too_expensive = too_expensive



import os
import sys
from pathlib import Path

from sphinx_codeautolink import clean_pycon

# Insert package root to path
_src_dir = Path(os.path.realpath(__file__)).parent
_package_root = _src_dir.parent.parent / "src"
sys.path.insert(0, str(_package_root))
sys.path.insert(0, str(_src_dir))

project = "sphinx-codeautolink"
author = "Felix Hildén"
copyright = "2021-2022, Felix Hildén"
version = Path(_package_root, "sphinx_codeautolink", "VERSION").read_text().strip()
release = version

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx.ext.extlinks",
    "sphinx_rtd_theme",
    "sphinx_codeautolink",
    "matplotlib.sphinxext.plot_directive",
    "sphinx.ext.doctest",
    "IPython.sphinxext.ipython_directive",
    "IPython.sphinxext.ipython_console_highlighting",
]

# Builtin options
html_theme = "sphinx_rtd_theme"
python_use_unqualified_type_names = True

# Extension options
codeautolink_autodoc_inject = True
codeautolink_custom_blocks = {"python3": None, "pycon3": clean_pycon}

autodoc_default_options = {
    "members": True,
    "undoc-members": True,
}
autodoc_typehints = "description"
intersphinx_mapping = {
    "python": ('https://docs.python.org/3/', None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
}
extlinks = {
    'issue': ('https://github.com/felix-hilden/sphinx-codeautolink/issues/%s', '#'),
}
# Copy plot directive options from Seaborn
# Include the example source for plots in API docs
plot_include_source = True
plot_formats = [("png", 90)]
plot_html_show_formats = False
plot_html_show_source_link = False


def setup(app):
    app.add_object_type(
        'confval',
        'confval',
        objname='configuration value',
        indextemplate='pair: %s; configuration value'
    )



import setuptools
import os
from pathlib import Path

root = Path(os.path.realpath(__file__)).parent
version_file = root / "src" / "sphinx_codeautolink" / "VERSION"
readme_file = root / "readme_pypi.rst"

setuptools.setup(
    name="sphinx-codeautolink",
    version=version_file.read_text().strip(),
    license="MIT",
    description="Automatic links from code examples to reference documentation.",
    keywords="sphinx extension code link",
    long_description=readme_file.read_text(),
    long_description_content_type="text/x-rst",

    url="https://pypi.org/project/sphinx-codeautolink",
    download_url="https://pypi.org/project/sphinx-codeautolink",
    project_urls={
        "Source": "https://github.com/felix-hilden/sphinx-codeautolink",
        "Issues": "https://github.com/felix-hilden/sphinx-codeautolink/issues",
        "Documentation": "https://sphinx-codeautolink.rtfd.org",
    },

    author="Felix Hildén",
    author_email="felix.hilden@gmail.com",
    maintainer="Felix Hildén",
    maintainer_email="felix.hilden@gmail.com",

    packages=setuptools.find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,

    python_requires=">=3.6",
    install_requires=[
        'sphinx>=3.2.0',
        'beautifulsoup4',
        'dataclasses;python_version<"3.7"',
    ],
    # Keep extras in sync with requirements manually
    extras_require={
        "ipython": ["ipython"],
    },

    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Sphinx',
        'Framework :: Sphinx :: Extension',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Documentation',
        'Topic :: Documentation :: Sphinx',
        'Topic :: Software Development :: Documentation',
    ],
)



from ._util import refs_equal


class TestChain:
    @refs_equal
    def test_attr_call_doesnt_contain_call(self):
        s = 'import a\na.attr()'
        refs = [('a', 'a'), ('a.attr', 'a.attr')]
        return s, refs

    @refs_equal
    def test_attr_call_attr_split_in_two(self):
        s = 'import a\na.attr().b'
        refs = [('a', 'a'), ('a.attr', 'a.attr'), ('a.attr.().b', 'b')]
        return s, refs

    @refs_equal
    def test_attr_call_attr_call_attr_split_in_three(self):
        s = 'import a\na.attr().b().c'
        refs = [
            ('a', 'a'),
            ('a.attr', 'a.attr'),
            ('a.attr.().b', 'b'),
            ('a.attr.().b.().c', 'c'),
        ]
        return s, refs
